#ifndef AR_UTILS
#define AR_UTILS
/*
 *  include file for arutils.c functions
 *  BACI archiver
 *     Bill Bynum
 *     May 2000
 */

#include <sys/types.h>
#include <stdio.h>
#include "artables.h"

char* new_string(int len);
/*
 *  create space for a string of length 'len' in the heap and
 *  return a pointer to it
 */

char* save_string(char* s);
/* 
 *  save a string on the heap 
 */

char* normalize(char* file);
/* 
 * local equivalent of 'basename', copied from GNU ar 
 * returns a pointer to the next char after the rightmost '/'
 * if there is one or a pointer to the beginning of the string
 */

void open_archive(Archive* ar, char* arfname);
/*
 * If archive file exists, opens it for reading and sets its file descriptor
 * Sets file descriptor to -1 if file doesn't exist.
 * Aborts for any other reason
 */
 
void read_archive(Archive* ar);
/*
 * if 'ar' is non-NULL and '*ar' has been opened for reading, 
 * then reads archive information into the '*ar' struct;
 * otherwise, does nothing
 */

void PositionToRead(Archive* ar, Member* m);
   /* position the read file descriptor so that member *m can be read */
 
void reopen_archive(Archive* ar, char* arfname);
/*
 * opens the archive file with 'open_archive'
 * aborts if archive file doesn't exist
 */

char* make_new_arfname(char* arfname);
/*
 *  creates a name for the new archive file 'new_arfname' that is similar 
 *  to the name of the old archive file 'arfname'
 */

void open_new_archive(Archive* ar, char* new_arfname);
/*
 *  opens for writing an archive file with 'new_arfname' 
 *  and stores file pointer to the file in *ar
 *  'new_arfname' is assumed to point to non-volatile storage!!!
 *  aborts if there is a problem
 */

int next_arstring(Archive* ar, char* buf, int len);
/*
 *  reads up to 'len' characters at the current file offset 
 *  in the file described by 'ar->rf' into the character buffer 'buf'.  
 *  This buffer is expected to be of length at least 'len' + 1, because 
 *  the string is terminated with a null byte as soon as either a blank is 
 *  encountered or 'len' characters are read.
 *  returns 0 on success, returns 1 on EOF, aborts otherwise
 */

int read_arstring(Archive* ar, off_t offset, char *buf,  int len);
/*
 *  reads up to 'len' characters at a file offset 'offset' in the file 
 *  in the file described by 'ar->rf' into the character buffer 'buf'.  
 *  is expected to be of length at least 'len' + 1, because the string 
 *  is terminated with a null byte as soon as either a blank is 
 *  encountered or 'len' characters are read.
 *  returns 0 on success, returns 1 on EOF, aborts otherwise
 */

long read_long(FILE* fd);
/* 
 *  reads a long at current offset of file described by 'fd' 
 *  or dies trying
 */

void printSymbol(Symbol* s);
   /* prints information about symbol to stdout */
/*
 *  $Log: arutils.h,v $
 *  Revision 1.9  2001/07/20 18:39:21  bynum
 *  add PositionToRead
 *
 *  Revision 1.8  2001/06/06 20:28:25  bynum
 *  DOS-ize include filename
 *
 *  Revision 1.7  2001/06/04 21:20:38  bynum
 *  add read_archive proc
 *
 *  Revision 1.6  2001/06/01 16:06:54  bynum
 *  add nonfatal
 *
 *  Revision 1.5  2000/05/30 15:50:56  bynum
 *  switch to BACI from ar68xx, delete fatal protoype (use BACI lib version),
 *  change nonfatal to varargs format, add printSymbol protoype
 *
 * Revision 1.4  1996/04/23  17:07:51  bynum
 * add address comment, change parm of 'read_long' to type FILE*
 *
 * Revision 1.3  1996/01/24  17:06:59  bynum
 * change to use of Archive struct
 *
 * Revision 1.2  1996/01/20  08:54:36  bynum
 * add 'normalize' and 'nonfatal'
 *
 * Revision 1.1  1996/01/15  13:37:45  bynum
 * Initial revision
 *
 */
#endif
